@@ -3,6 +3,8 @@ package Lingy::Reader;
 
 use Lingy::Common;
 
+our $feature_keyword = ':lingy.pl';
+
 my $tokenize_re = qr/
     (?:                     # Ignore:
         \#\!.* |                # hashbang line
@@ -10,7 +12,11 @@ my $tokenize_re = qr/
         ;.*                     # comments
     )*
     (                       # Capture all these tokens:
+        \#\_ |                  # Ignore next form
+        \#\' |                  # Var
         \#\( |                  # Lambda
+        \#\{ |                  # HashSet
+        \#\? |                  # Reader conditional
         ~@ |                    # Unquote-splice token
         [\[\]{}()'`~^@] |       # Single character tokens
         \#?                     # Possibly a regex
@@ -28,6 +34,7 @@ sub new {
     bless {
         tokens => [],
         lambda => [],
+        ignore => 0,
         @_,
     }, $class;
 }
@@ -55,17 +62,23 @@ sub read_str {
 
 sub read_form {
     my ($self) = @_;
-    local $_ = $self->{tokens}[0];
+    my $tokens = $self->{tokens};
+    while (not @$tokens) { $self->read_more }
+    local $_ = $tokens->[0];
     /^\($/ ? $self->read_list(LIST, ')') :
     /^\[$/ ? $self->read_list(VECTOR, ']') :
-    /^\{$/ ? $self->read_hash_map(HASHMAP, '}') :
+    /^\{$/ ? $self->read_hash('map') :
+    /^\#\{$/ ? $self->read_hash('set') :
     /^'$/ ? $self->read_quote('quote') :
     /^`$/ ? $self->read_quote('quasiquote') :
     /^~$/ ? $self->read_quote('unquote') :
     /^~\@$/ ? $self->read_quote('splice-unquote') :
     /^\@$/ ? $self->read_quote('deref') :
     /^\^$/ ? $self->with_meta :
-    /^#\(/ ? $self->read_lambda :
+    /^#\'$/ ? $self->read_var :
+    /^#\($/ ? $self->read_lambda :
+    /^#_$/ ? $self->read_ignore :
+    /^#\?$/ ? $self->read_cond :
     /^%\d*$/ ? $self->read_lambda_symbol :
     $self->read_scalar;
 }
@@ -100,6 +113,16 @@ sub read_list {
     }
 }
 
+sub read_var {
+    my ($self) = @_;
+    shift @{$self->{tokens}};
+    my $var = shift @{$self->{tokens}};
+    if ($var !~ m</>) {
+        $var = RT->current_ns_name . "/$var";
+    }
+    VAR->new($var);
+}
+
 sub read_lambda {
     my ($self) = @_;
     my $tokens = $self->{tokens};
@@ -145,32 +168,70 @@ sub read_lambda_symbol {
     $lambda->{sym}{$token} //= symbol("p${num}_${\RT->nextID}");
 }
 
-sub read_hash_map {
-    my ($self, $type, $end) = @_;
+sub read_hash {
+    my ($self, $type) = @_;
     my $tokens = $self->{tokens};
     shift @$tokens;
-    my $hash = $type->new([]);
-    my $i = 0;
-    my $key;
+    my $hash = [];
     while (1) {
         while (@$tokens > 0) {
-            if ($tokens->[0] eq $end) {
+            if ($tokens->[0] eq '}') {
+                shift @$tokens;
+                my $method = "make_hash_$type";
+                return $self->$method($hash);
+            }
+            push @$hash, $self->read_form;
+        }
+        $self->read_more and next;
+        err "Reached end of input in 'read_hash'";
+    }
+}
+
+sub make_hash_map {
+    my ($self, $data) = @_;
+    err "Map literal must contain an even number of forms"
+        if @$data % 2 and not $self->{ignore};
+    my %hash;
+    for (my $i = 0; $i < @$data; $i += 2) {
+        err "Duplicate key: '$data->[$i]'"
+            if $hash{$data->[$i]}++ and
+                not $self->{ignore};
+    }
+    HASHMAP->new($data);
+}
+
+sub make_hash_set {
+    my ($self, $data) = @_;
+    my %hash;
+    my $set = [];
+    for my $elem (@$data) {
+        err "Duplicate key: '$elem'"
+            if $hash{$elem}++ and
+                not $self->{ignore};
+        push @$set, $elem, $elem;
+    }
+    HASHSET->new($data);
+}
+
+sub read_hash_set {
+    my ($self) = @_;
+    my $tokens = $self->{tokens};
+    shift @$tokens;
+    my $hash = HASHSET->new([]);
+    my $i = 0;
+    while (1) {
+        while (@$tokens > 0) {
+            if ($tokens->[0] eq '}') {
                 shift @$tokens;
-                err "Map literal must contain an even number of forms"
-                    if defined $key;
                 return $hash;
             }
             $i++;
-            if ($i % 2) {
-                $key = $self->read_form;
-            } else {
-                my $key_str = $type->_get_key($key);
-                err "Duplicate key: '$key'"
-                    if exists $hash->{$key_str};
-                my $val= $self->read_form;
-                $hash->{$key_str} = $val;
-                undef $key;
-            }
+            my $val = $self->read_form;
+            my $key = HASHSET->_get_key($val);
+            err "Duplicate key: '$val'"
+                if exists $hash->{$key} and
+                    not $self->{ignore};
+            $hash->{$key} = $val;
         }
         $self->read_more and next;
         err "Reached end of input in 'read_hash_map'";
@@ -195,7 +256,20 @@ my $regexp_unescape = {
     'w' => "\\w",
     'z' => "\\z",
     '"' => '\\"',
+    '.' => '\\.',
     '\\' => "\\\\",
+    '[' => "\\[",
+    ']' => "\\]",
+    '{' => "\\{",
+    '}' => "\\}",
+    '(' => "\\(",
+    ')' => "\\)",
+    '+' => "\\+",
+    '*' => "\\*",
+    '?' => "\\?",
+    '^' => "\\^",
+    '$' => "\\\$",
+    '|' => "\\|",
 };
 sub read_scalar {
     my ($self) = @_;
@@ -205,16 +279,18 @@ sub read_scalar {
         my $is_regex = $1;
         if (/^$string_re$/) {
             s/^$string_re$/$1/;
-            if ($is_regex) {
-                s/\\(.)/
-                    $regexp_unescape->{$1}
-                        or err("Unsupported escape character '\\$1'")
-                /ge;
-            } else {
-                s/\\(.)/
-                    $string_unescape->{$1}
-                        or err("Unsupported escape character '\\$1'")
-                /ge;
+            if (not $self->{ignore}) {
+                if ($is_regex) {
+                    s/\\(.)/
+                        $regexp_unescape->{$1}
+                            or err("Unsupported escape character '\\$1'")
+                    /ge;
+                } else {
+                    s/\\(.)/
+                        $string_unescape->{$1}
+                            or err("Unsupported escape character '\\$1'")
+                    /ge;
+                }
             }
             return $is_regex ? REGEX->new($_) : STRING->new($_);
         }
@@ -231,7 +307,7 @@ sub read_scalar {
     return true if $_ eq 'true';
     return false if $_ eq 'false';
     return KEYWORD->new($_) if /^:/;
-    return NUMBER->new($_) if /^-?\d+$/;
+    return NUMBER->new($_) if /^-?\d+(?:\.\d+)?$/;
     return CHARACTER->read($_) if /^\\/;
     err "Unmatched delimiter: '$_'" if /^[\)\]\}]$/;
     return $self->read_symbol($_);
@@ -264,17 +340,65 @@ sub read_quote {
 }
 
 sub with_meta {
-    my ($self, $quote) = @_;
-    shift @{$self->{tokens}};
+    my ($self) = @_;
+
+    my $tokens = $self->{tokens};
+    my @meta;
+
+    while (@$tokens > 2 and $tokens->[0] eq '^') {
+        shift @$tokens;
+        my $meta = $self->read_form;
+        my $type = ref($meta);
+        if ($type eq SYMBOL or $type eq STRING) {
+            unshift @meta, KEYWORD->new(':tag'), $meta;
+        } elsif ($type eq KEYWORD) {
+            unshift @meta, $meta, true;
+        } elsif ($type eq HASHMAP) {
+            unshift @meta, %$meta;
+        } else {
+            err "Metadata must be Symbol,Keyword,String or Map"
+        }
+    }
+
+    my $meta = $self->make_hash_map(\@meta);
 
-    my $meta = $self->read_form;
     my $form = $self->read_form;
 
-    bless [
-        symbol('with-meta'),
-        $form,
-        $meta,
-    ], LIST;
+    RT->meta->{"$form"} = $meta;
+
+    return $form;
+}
+
+sub read_ignore {
+    my ($self) = @_;
+    $self->{ignore}++;
+    my $tokens = $self->{tokens};
+    shift @$tokens;
+    while (not defined $self->read_form) {}
+    $self->{ignore}--;
+    return;
+}
+
+sub read_cond {
+    my ($self) = @_;
+    my $tokens = $self->{tokens};
+    shift @$tokens;
+    err "read-cond body must be a list"
+        unless @$tokens and $tokens->[0] eq '(';
+    my $list = $self->read_form;
+    err "read-cond requires an even number of forms"
+        if @$list % 2;
+    while (@$list) {
+        my ($keyword, $form) = splice(@$list, 0, 2);
+        err "Feature should be a keyword: $keyword"
+            unless $keyword->isa(KEYWORD);
+        if ("$keyword" eq $feature_keyword or
+            "$keyword" eq ':default'
+        ) {
+            return $form;
+        }
+    }
+    return;
 }
 
 1;
